import configparser
import pandas as pd
import logging

from datetime import datetime
from time import sleep

from alphatools.utils.smartapi_helper import SmartApiHelper


class BackTestingApp:
    logger = logging.getLogger(__name__)
    instruments_list = []
    start_date = None
    end_date = None
    data_interval = 'ONE_MINUTE'

    def __init__(self, config_file):
        """

        :param config_file:
        """
        super().__init__()
        self.candle_info_df = None
        cfg_parser = configparser.ConfigParser()
        cfg_parser.read(config_file)
        self.api_key = cfg_parser.get('SMARTAPI_LOGIN', 'API_KEY')
        self.client_code = cfg_parser.get('SMARTAPI_LOGIN', 'CLIENT_CODE')
        self.password = cfg_parser.get('SMARTAPI_LOGIN', 'PASSWORD')
        self.totp_key = cfg_parser.get('SMARTAPI_LOGIN', 'TOTP_KEY')

    @staticmethod
    def _get_time(time):
        return datetime.strptime(time, '%Y-%m-%dT%H:%M:%S%z')

    def on_md(self, data_row):
        """
        Basic onMd template. You can create a child class with backtestingApp as parent to
        come up with your own onMd function. See basic documentation.
        :param data_row:
        :return:
        """
        self.logger.info("Received row: {}".format(data_row))
        pass

    def add_instrument(self, token, exchange):
        """
        Adds an instrument to the subscriber list
        :param token: token number of the instrument
        :param exchange: Exchange in which the instrument is traded
        :return:
        """
        self.instruments_list.append((token, exchange))

    def set_start_date(self, start_date):
        """
        Start datetime of simulation
        :param start_date: datetime
        :return: None
        """

        self.start_date = start_date

    def set_end_date(self, end_date):
        """
        End datetime of simulation
        :param end_date: datetime
        :return: None
        """
        self.end_date = end_date

    def set_interval(self, interval='ONE_MINUTE'):
        """
        Set frequency of requested data for simulation. Possible values:
                Interval	Description
                ONE_MINUTE	1 Minute
                THREE_MINUTE	3 Minute
                FIVE_MINUTE	5 Minute
                TEN_MINUTE	10 Minute
                FIFTEEN_MINUTE	15 Minute
                THIRTY_MINUTE	30 Minute
                ONE_HOUR	1 Hour
                ONE_DAY	1 Day

        :param interval: Requested interval

        :return:
        """
        self.data_interval = interval

    def load_data(self):
        """
        Based on the instrument info requested, sends get Request to AB smart API server

        :return: None
        """
        self.candle_info_df = pd.DataFrame.from_records([],
                                                        columns=['Timestamp', 'OPEN', 'HIGH', 'LOW', 'CLOSE', 'VOLUME'])

        for _date in pd.date_range(self.start_date, self.end_date):
            for token, exchange in self.instruments_list:
                candle_info_params = {
                    "exchange": exchange,
                    "symboltoken": token,
                    "interval": self.data_interval,
                    "fromdate": datetime.strftime(_date, '%Y-%m-%d 00:00'),
                    "todate": datetime.strftime(_date, '%Y-%m-%d 23:59')
                }
                self.logger.info("Sending candle request for {}".format(candle_info_params))
                api_helper = SmartApiHelper(self.api_key, self.client_code, self.password, self.totp_key)
                results = api_helper.get_candle_info(candle_info_params)['data']
                if not results:
                    self.logger.error("No data available for params: {}".format(candle_info_params))
                    continue
                df = pd.DataFrame.from_records(results,
                                               columns=['Timestamp', 'OPEN', 'HIGH', 'LOW', 'CLOSE', 'VOLUME'])
                df['Token'] = token
                df['Timestamp'] = df['Timestamp'].apply(lambda x: self._get_time(x))
                self.candle_info_df = pd.concat([self.candle_info_df, df])
                sleep(0.35)

        self.candle_info_df = self.candle_info_df.sort_values(by=['Timestamp', 'Token']).reset_index(drop=True)

    def get_candle_info_df(self, num_rows=-1):
        """
        Returns top n rows of the candle info dataframe

        :param num_rows: Number of rows required for training purposes
        :return: pandas dataframe
        """
        if num_rows == -1:
            return self.candle_info_df

        return self.candle_info_df.head(num_rows)

    def get_row_number(self, timestamp):
        """
        Returns index of the last row of df such that df.timestamp < timestamp.
        Makes it more convenient to split training and test dataset.

        For example, you want to train the model initially for say 9:30-11:30 and
        run the rest of the simulation online. Call this function to get the row number
        corresponding to 11:30 and send it to the functions simulate and
        get_candle_info_df() function.

        :param timestamp:  Benchmark time to split into training and test
        :return: Index of row
        """
        return len(self.candle_info_df[self.candle_info_df['Timestamp'] < timestamp])

    def simulate(self, start=0):
        """
        Runs simulation from the required position. It is recommended to use simulate without
        args to simulate everything online one by one.

        :param start: Offset starting position of online simulation
        :return: None
        """
        for idx, row in self.candle_info_df.iterrows():
            if idx >= start:
                self.on_md(row)



import logging
from collections import defaultdict

from multipledispatch import dispatch

from alphatools.utils.smartapi_helper import SmartApiHelper


class TokenManager:
    logger = logging.getLogger(__name__)
    token_to_instrument_info = defaultdict(None)
    symbol_to_instrument_info = defaultdict(None)

    def __init__(self):
        instruments_json = SmartApiHelper.get_instruments_list()
        self.logger.info("Instruments Received: {}".format(len(instruments_json)))
        for instrument in instruments_json:
            try:
                token = int(instrument['token'])
                symbol = instrument['symbol']
                self.token_to_instrument_info[token] = instrument
                self.symbol_to_instrument_info[symbol] = instrument
            except ValueError as e:
                self.logger.debug('Invalid token for instrument {}. Skipping this row'.format(instrument))

        self.logger.info("Instruments processed successfully: {}".format(len(self.token_to_instrument_info)))


    @dispatch(int)
    def get_instrument(self, token):
        """
        :param token: Token number of the instrument specified by NSE
        :return: If token is valid {
            'token': '48756',
            'symbol': 'BANKNIFTY23FEB23FUT',
            'name': 'BANKNIFTY',
            'expiry': '23FEB2023',
            'strike': '-1.000000',
            'lotsize': '25',
            'instrumenttype': 'OPTIDX',
            'exch_seg': 'NFO',
            'tick_size': '5.000000'
        }
        Else None
        """
        return self.token_to_instrument_info.get(token)

    @dispatch(str)
    def get_instrument(self, symbol):
        return self.symbol_to_instrument_info.get(symbol)

    def get_fut(self, name, expiry):
        """
        Returns properties of future with requested params
        :param name: Underlying symbol
        :param expiry: Expiry Date
        :return: Refer @get_instrument for more details
        """
        symbol = "{}{}FUT".format(name, expiry)
        self.logger.info("Searching for symbol: {}".format(symbol))
        return self.get_instrument(symbol)

    def get_opt(self, name, strike, expiry, option_type):
        """
        Returns properties of option with requested params
        :param name: Underlying symbol
        :param strike: Strike Price
        :param expiry: Expiry Date
        :param option_type: Put or call
        :return: Returns a dictionary in the following format or None if the instrument is invalid/not found
        """
        symbol = "{}{}{}{}".format(name, strike, expiry, option_type)
        self.logger.info("Searching for symbol: {}".format(symbol))
        return self.get_instrument(symbol)



import datetime
from unittest import TestCase
from alphatools import backtesting_app
from datetime import datetime

from alphatools.utils.token_manager import TokenManager
import logging

logging.basicConfig(level=logging.INFO)


class TestBackTestingApp(TestCase):
    logger = logging.getLogger(__name__)

    def test_run_backtesting_engine_basic(self):
        app = backtesting_app.BackTestingApp("/Users/jaskiratsingh/projects/smart-api-creds.ini")
        app.set_start_date(datetime.strptime('2022-12-20 11:39:00+05:30', '%Y-%m-%d %H:%M:%S%z'))
        app.set_end_date(datetime.strptime('2022-12-20 11:39:00+05:30', '%Y-%m-%d %H:%M:%S%z'))
        app.set_interval('ONE_MINUTE')
        app.add_instrument(53825, "NFO")
        app.add_instrument(48756, "NFO")
        app.load_data()                     # Loads the data into a dataframe
        app.simulate()                      # Starts simulation from the beginning
        app.get_row_number(datetime.strptime('2022-12-23 11:39:00+05:30', '%Y-%m-%d %H:%M:%S%z'))
        app.get_candle_info_df()            # Returns the entire simulation dataframe

    def test_run_backtesting_engine_advanced(self):
        app = backtesting_app.BackTestingApp("/Users/jaskiratsingh/projects/smart-api-creds.ini")
        app.set_start_date(datetime.strptime('2022-12-20 11:39:00+05:30', '%Y-%m-%d %H:%M:%S%z'))
        app.set_end_date(datetime.strptime('2022-12-20 11:39:00+05:30', '%Y-%m-%d %H:%M:%S%z'))
        app.set_interval('ONE_MINUTE')
        app.add_instrument(53825, "NFO")
        app.add_instrument(48756, "NFO")
        app.load_data()                     # Loads the data into a dataframe
        app.get_candle_info_df(-1)          # Returns entire candle info dataframe
        app.get_candle_info_df(2)           # Returns first two rows of the dataframe
        app.simulate(3)                     # Starts simulating from index 3

    def test_run_token_manager(self):
        app = TokenManager()
        self.logger.info(app.get_instrument('BANKNIFTY23FEB23FUT'))
        self.logger.info(app.get_instrument(48756))



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open(".github/VERSION", "r") as fh:
    app_version = fh.read()

setuptools.setup(
    name="alphatools_jv",
    version=app_version,
    author="Jaskirat Singh",
    author_email="jaskiratsingh1208@gmail.com",
    description="A small example package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jaskirat1208/backtest-platform",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
)



from smartapi import SmartWebSocket, SmartConnect
import mintotp
import requests
import logging

INSTRUMENT_API_URL = 'http://margincalculator.angelbroking.com/OpenAPI_File/files/OpenAPIScripMaster.json'


class SmartApiHelper:
    logger = logging.getLogger(__name__)

    def __init__(self, api_key, client_code, password, totp_key):
        self.api_key = api_key
        self.client_code = client_code
        self.passwd = password
        self.totp_key = totp_key

    @staticmethod
    def get_instruments_list():
        instruments_csv_list = requests.get(INSTRUMENT_API_URL)
        return instruments_csv_list.json()

    def get_candle_info(self, candle_info_params):
        candle_info_results = []
        try:
            smart_conn = SmartConnect(api_key=self.api_key)
            totp = mintotp.totp(self.totp_key)
            smart_conn.generateSession(self.client_code, self.passwd, totp)
            candle_info_results = smart_conn.getCandleData(candle_info_params)
            smart_conn.terminateSession(self.client_code)
        except Exception as e:
            self.logger.error("Historical API failed with exception: {}".format(e.message))

        return candle_info_results
